from dataclasses import dataclass

import torch
from torch import nn
import torch.nn.functional as F

from nn_zoo.models.components import SelfAttention

__all__ = ["GPTConfig", "GPT"]


@dataclass
class GPTConfig:
    vocab_size: int
    block_size: int
    emb_size: int
    heads: int
    num_layers: int
    attn_dropout: float
    ff_mult: int
    ff_dropout: float


class MLP(nn.Module):
    def __init__(self, config: GPTConfig):
        super().__init__()
        self.config = config

        self.fc1 = nn.Linear(config.emb_size, config.ff_mult * config.emb_size)
        self.fc2 = nn.Linear(config.ff_mult * config.emb_size, config.emb_size)
        self.act = nn.GELU()

        if config.ff_dropout > 0:
            self.dropout = nn.Dropout(config.ff_dropout)

    def forward(self, x):
        x = self.fc1(x)
        x = self.act(x)
        x = self.fc2(x)

        if hasattr(self, "dropout"):
            x = self.dropout(x)

        return x


class Block(nn.Module):
    def __init__(self, config: GPTConfig):
        super().__init__()
        self.ln_1 = nn.LayerNorm(config.emb_size)
        self.attn = SelfAttention(config.emb_size, config.heads, config.attn_dropout)
        self.ln_2 = nn.LayerNorm(config.emb_size)
        self.mlp = MLP(config)

    def forward(self, x):
        x = x + self.attn(self.ln_1(x), True)
        x = x + self.mlp(self.ln_2(x))
        return x


class GPT(nn.Module):
    def __init__(self, config: GPTConfig):
        super().__init__()
        self.config = config

        self.token_emb = nn.Embedding(config.vocab_size, config.emb_size)
        self.pos_emb = nn.Embedding(config.block_size, config.emb_size)

        self.blocks = nn.ModuleList([Block(config) for _ in range(config.num_layers)])

        self.ln = nn.LayerNorm(config.emb_size)
        self.head = nn.Linear(config.emb_size, config.vocab_size, bias=False)

        # tie weights
        self.head.weight = self.token_emb.weight

        # initialize weights
        self.apply(self._init_weights)

    def _init_weights(self, module):
        if isinstance(module, nn.Linear):
            nn.init.normal_(module.weight, std=0.02)
            if module.bias is not None:
                nn.init.zeros_(module.bias)

        elif isinstance(module, nn.Embedding):
            nn.init.normal_(module.weight, std=0.02)

        elif isinstance(module, nn.LayerNorm):
            nn.init.ones_(module.weight)
            nn.init.zeros_(module.bias)

    def forward(self, x):
        B, T = x.size()
        assert (
            not T > self.config.block_size
        ), "Sequence length is longer than block size"

        emb = self.token_emb(x)
        pe = self.pos_emb(torch.arange(T - 1, -1, step=-1))

        x = emb + pe

        for block in self.blocks:
            x = block(x)

        x = self.ln(x)
        return self.head(x)

    def loss(self, y, y_pred):
        # Input is a contiguous tensor
        y = y.flatten()
        y_pred = y_pred.view(-1, y_pred.size(-1))

        return F.cross_entropy(y_pred, y)

    def get_param_count(self):
        return sum(p.numel() for p in self.parameters())

    @torch.no_grad()
    def generate(
        self,
        start_seq: list[int],
        max_len: int = 128,
        temperature: float = 1.0,
        top_k: int | None = None,
    ):
        if top_k is None:
            top_k = self.config.vocab_size
        self.eval()

        generated = start_seq
        primer_t = torch.as_tensor(start_seq).unsqueeze(0)

        for _ in range(max_len):
            if primer_t.size(1) >= self.config.block_size:
                primer_t = primer_t[:, -self.config.block_size :]

            out = self(primer_t)
            out = out[:, -1, :] / temperature
            out = F.softmax(out, dim=-1)
            out = torch.topk(out, top_k, dim=-1)[0]
            out = torch.multinomial(out, num_samples=1)

            gen = out.item()

            generated.append(int(gen))

            primer_t = torch.cat((primer_t, out), dim=1)

        return generated


if __name__ == "__main__":
    config = GPTConfig(
        vocab_size=32,
        block_size=4,
        emb_size=8,
        heads=1,
        num_layers=1,
        attn_dropout=0.1,
        ff_mult=4,
        ff_dropout=0.1,
    )

    model = GPT(config)
    print(model.get_param_count())
    print(model.generate([0, 1, 2]))



import torch
from nn_zoo.models import VQVAE


def test_vqvae():
    model = VQVAE(
        in_channels=3,
        out_channels=3,
        num_hiddens=32,
        num_downsampling_layers=2,
        num_residual_layers=2,
        num_residual_hiddens=32,
        embedding_dim=64,
        num_embeddings=512,
        use_ema=True,
        decay=0.99,
        epsilon=1e-5,
    )
    assert model
    assert model.encoder
    assert model.decoder

    x = torch.rand(1, 3, 32, 32)
    y = model(x)
    assert y["x_recon"].shape == (1, 3, 32, 32)


if __name__ == "__main__":
    test_vqvae()



import torch
from nn_zoo.models.components import SelfAttention


def test_self_attention():
    x = torch.randn(1, 32, 64)
    attn = SelfAttention(64, 8)
    assert attn(x).shape == (1, 32, 64)


if __name__ == "__main__":
    test_self_attention()



from .attention import SelfAttention
from .residual import ResidualBasicBlock, ResidualBottleNeckBlock, ResidualStack
from .vq import VectorQuantizer

__all__ = [
    "SelfAttention",
    "ResidualBasicBlock",
    "ResidualBottleNeckBlock",
    "ResidualStack",
    "VectorQuantizer",
]



import torch.nn as nn
import torch.nn.functional as F
from nn_zoo.models.components import VectorQuantizer
from nn_zoo.models.components import ResidualStack

__all__ = ["VQVAE"]


class Encoder(nn.Module):
    def __init__(
        self,
        in_channels,
        num_hiddens,
        num_downsampling_layers,
        num_residual_layers,
        num_residual_hiddens,
    ):
        super().__init__()
        # See Section 4.1 of "Neural Discrete Representation Learning".
        # The last ReLU from the Sonnet example is omitted because ResidualStack starts
        # off with a ReLU.
        conv = nn.Sequential()
        for downsampling_layer in range(num_downsampling_layers):
            if downsampling_layer == 0:
                out_channels = num_hiddens // 2
            elif downsampling_layer == 1:
                (in_channels, out_channels) = (num_hiddens // 2, num_hiddens)

            else:
                (in_channels, out_channels) = (num_hiddens, num_hiddens)

            conv.add_module(
                f"down{downsampling_layer}",
                nn.Conv2d(
                    in_channels=in_channels,
                    out_channels=out_channels,
                    kernel_size=4,
                    stride=2,
                    padding=1,
                ),
            )
            conv.add_module(f"relu{downsampling_layer}", nn.ReLU())

        conv.add_module(
            "final_conv",
            nn.Conv2d(
                in_channels=num_hiddens,
                out_channels=num_hiddens,
                kernel_size=3,
                padding=1,
            ),
        )
        self.conv = conv
        self.residual_stack = ResidualStack(
            n_blocks=num_residual_layers,
            in_channels=num_hiddens,
            out_channels=num_hiddens,
            kernel_size=3,
        )

    def forward(self, x):
        h = self.conv(x)
        return self.residual_stack(h)


class Decoder(nn.Module):
    def __init__(
        self,
        embedding_dim,
        num_hiddens,
        num_upsampling_layers,
        num_residual_layers,
        n_channels,
        out,
    ):
        super().__init__()
        # See Section 4.1 of "Neural Discrete Representation Learning".
        self.conv = nn.Conv2d(
            in_channels=embedding_dim,
            out_channels=num_hiddens,
            kernel_size=3,
            padding=1,
        )
        self.residual_stack = ResidualStack(
            n_blocks=num_residual_layers,
            in_channels=num_hiddens,
            out_channels=num_hiddens,
            kernel_size=3,
        )
        upconv = nn.Sequential()
        for upsampling_layer in range(num_upsampling_layers):
            if upsampling_layer < num_upsampling_layers - 2:
                (in_channels, out_channels) = (num_hiddens, num_hiddens)

            elif upsampling_layer == num_upsampling_layers - 2:
                (in_channels, out_channels) = (num_hiddens, num_hiddens // 2)

            elif upsampling_layer > num_upsampling_layers - 2:
                (in_channels, out_channels) = (num_hiddens // 2, out)
            else:
                raise ValueError(
                    f"Invalid upsampling layer: {upsampling_layer}. In Encoder."
                )

            upconv.add_module(
                f"up{upsampling_layer}",
                nn.ConvTranspose2d(
                    in_channels=in_channels,
                    out_channels=out_channels,
                    kernel_size=4,
                    stride=2,
                    padding=1,
                ),
            )
            if upsampling_layer < num_upsampling_layers - 1:
                upconv.add_module(f"relu{upsampling_layer}", nn.ReLU())

        self.upconv = upconv

    def forward(self, x):
        h = self.conv(x)
        h = self.residual_stack(h)
        x_recon = self.upconv(h)
        return F.sigmoid(x_recon)


class VQVAE(nn.Module):
    def __init__(
        self,
        in_channels,
        out_channels,
        num_hiddens,
        num_downsampling_layers,
        num_residual_layers,
        num_residual_hiddens,
        embedding_dim,
        num_embeddings,
        use_ema,
        decay,
        epsilon,
    ):
        super().__init__()
        self.encoder = Encoder(
            in_channels,
            num_hiddens,
            num_downsampling_layers,
            num_residual_layers,
            num_residual_hiddens,
        )
        self.pre_vq_conv = nn.Conv2d(
            in_channels=num_hiddens, out_channels=embedding_dim, kernel_size=1
        )
        self.vq = VectorQuantizer(
            embedding_dim, num_embeddings, use_ema, decay, epsilon
        )
        self.decoder = Decoder(
            embedding_dim,
            num_hiddens,
            num_downsampling_layers,
            num_residual_layers,
            num_residual_hiddens,
            out_channels,
        )

    def quantize(self, x):
        z = self.pre_vq_conv(self.encoder(x))
        (z_quantized, dictionary_loss, commitment_loss, encoding_indices) = self.vq(z)
        return (z_quantized, dictionary_loss, commitment_loss, encoding_indices)

    def forward(self, x):
        (z_quantized, dictionary_loss, commitment_loss, encoding_indices) = (
            self.quantize(x)
        )
        x_recon = self.decoder(z_quantized)
        return {
            "dictionary_loss": dictionary_loss,
            "commitment_loss": commitment_loss,
            "x_recon": x_recon,
            "encoding_indices": encoding_indices,
        }



import torch
import torch.optim as optim
import torch.nn.functional as F
import torchvision
import wandb
from lightning import pytorch as pl

from nn_zoo.models import VQVAE

__all__ = ["VQVAE_Trainer"]


class VQVAE_Trainer(pl.LightningModule):
    def __init__(
        self,
        sample_size=64,
        in_channels=1,
        out_channels=1,
        num_hiddens=64,
        num_downsampling_layers=4,
        num_residual_layers=1,
        num_residual_hiddens=128,
        embedding_dim=32,  # 32, 64, 128, 256
        num_embeddings=512,  # 256, 512, 1024, 2048
        use_ema=True,
        decay=0.99,
        epsilon=1e-5,
        beta=0.25,
        lr=2e-4,
        weight_decay=0.01,
        fid_features=2048,
        batch_size=64,
        dataset="mnist",
    ):
        super(VQVAE_Trainer, self).__init__()
        self.model = VQVAE(
            in_channels=in_channels,
            out_channels=out_channels,
            num_hiddens=num_hiddens,
            num_downsampling_layers=num_downsampling_layers,
            num_residual_layers=num_residual_layers,
            num_residual_hiddens=num_residual_hiddens,
            embedding_dim=embedding_dim,
            num_embeddings=num_embeddings,
            use_ema=use_ema,
            decay=decay,
            epsilon=epsilon,
        )

        self.beta = beta
        self.save_hyperparameters()

    def forward(self, x):
        return self.model(x)

    def training_step(self, batch, batch_idx):
        x, _ = batch

        out = self.model(x)
        recon_error = F.mse_loss(out["x_recon"], x)

        loss = recon_error + self.beta * out["commitment_loss"]

        if out["dictionary_loss"] is not None:
            loss += out["dictionary_loss"]
            self.log("train_dictionary_loss", out["dictionary_loss"])

        self.log("train/loss", loss)
        self.log("train/recon_error", recon_error)
        self.log("train/commitment_loss", out["commitment_loss"])

        return loss

    def validation_step(self, batch, batch_idx):
        x, _ = batch

        out = self.model(x)

        recon_error = F.mse_loss(out["x_recon"], x)

        loss = recon_error + self.beta * out["commitment_loss"]

        if out["dictionary_loss"] is not None:
            loss += out["dictionary_loss"]
            self.log("val_dictionary_loss", out["dictionary_loss"])

        self.log("val/loss", loss)
        self.log("val/recon_error", recon_error)
        self.log("val/commitment_loss", out["commitment_loss"])

        if batch_idx == 0:
            if self.global_step == 0 and batch_idx == 0:
                self.logger.experiment.log(
                    {
                        "original": wandb.Image(
                            torchvision.utils.make_grid(x[:64], nrow=8),
                            caption="Real Image",
                        )
                    }
                )

            self.logger.experiment.log(
                {
                    "reconstructed": wandb.Image(
                        torchvision.utils.make_grid(out["x_recon"][:64], nrow=8),
                        caption=f"Step {self.global_step}",
                    )
                }
            )

    def test_step(self, batch, batch_idx):
        x, _ = batch

        out = self.model(x)

        # Resize to 299x299
        x = F.interpolate(x, size=299)
        x_hat = F.interpolate(out["x_recon"], size=299)

        if x.shape[1] == 1:
            x = x.repeat(1, 3, 1, 1)
            x_hat = x_hat.repeat(1, 3, 1, 1)

        # Convert to uint8
        x = (x * 255).to(torch.uint8).cpu()
        x_hat = (x_hat * 255).to(torch.uint8).cpu()

        # Compute FID
        self.fid.update(x, real=True)
        self.fid.update(x_hat, real=False)

        fid_score = self.fid.compute()
        self.log("fid_score", fid_score)

    def configure_optimizers(self):
        return optim.AdamW(
            self.model.parameters(),
            lr=self.hparams.lr,
            amsgrad=True,
            weight_decay=self.hparams.weight_decay,
        )

    def train_dataloader(self):
        if self.hparams.dataset == "mnist":
            transform = torchvision.transforms.Compose(
                [
                    torchvision.transforms.Resize(self.hparams.sample_size),
                    (
                        torchvision.transforms.Grayscale()
                        if self.hparams.in_channels == 1
                        else torchvision.transforms.Lambda(lambda x: x)
                    ),
                    torchvision.transforms.ToTensor(),
                ]
            )
            dataset = torchvision.datasets.MNIST(
                root="data/mnist", train=True, transform=transform, download=True
            )

        elif self.hparams.dataset == "cifar10":
            transform = torchvision.transforms.Compose(
                [
                    torchvision.transforms.Resize(self.hparams.sample_size),
                    torchvision.transforms.ToTensor(),
                ]
            )
            dataset = torchvision.datasets.CIFAR10(
                root="data/cifar10", train=True, transform=transform, download=True
            )

        elif self.hparams.dataset == "celeba_hq":
            transform = torchvision.transforms.Compose(
                [
                    torchvision.transforms.Resize(self.hparams.sample_size),
                    torchvision.transforms.ToTensor(),
                ]
            )
            dataset = torchvision.datasets.ImageFolder(
                "data/celeba_hq/train", transform=transform
            )

        else:
            raise ValueError(f"Unknown dataset: {self.hparams.dataset}")

        return torch.utils.data.DataLoader(
            dataset,
            batch_size=self.hparams.batch_size,
            shuffle=True,
            num_workers=4,
            pin_memory=True,
            persistent_workers=True,
        )

    def val_dataloader(self):
        if self.hparams.dataset == "mnist":
            transform = torchvision.transforms.Compose(
                [
                    torchvision.transforms.Resize(self.hparams.sample_size),
                    torchvision.transforms.ToTensor(),
                ]
            )
            dataset = torchvision.datasets.MNIST(
                root="data/mnist", train=False, transform=transform, download=True
            )

        elif self.hparams.dataset == "cifar10":
            transform = torchvision.transforms.Compose(
                [
                    torchvision.transforms.Resize(self.hparams.sample_size),
                    torchvision.transforms.ToTensor(),
                ]
            )
            dataset = torchvision.datasets.CIFAR10(
                root="data/cifar10", train=False, transform=transform, download=True
            )

        elif self.hparams.dataset == "celeba_hq":
            transform = torchvision.transforms.Compose(
                [
                    torchvision.transforms.Resize(self.hparams.sample_size),
                    torchvision.transforms.ToTensor(),
                ]
            )
            dataset = torchvision.datasets.ImageFolder(
                "data/celeba_hq/val", transform=transform
            )

        else:
            raise ValueError(f"Unknown dataset: {self.hparams.dataset}")

        return torch.utils.data.DataLoader(
            dataset,
            batch_size=self.hparams.batch_size,
            num_workers=4,
            pin_memory=True,
            persistent_workers=True,
        )

    def test_dataloader(self):
        if self.hparams.dataset == "mnist":
            transform = torchvision.transforms.Compose(
                [
                    torchvision.transforms.Resize(self.hparams.sample_size),
                    torchvision.transforms.ToTensor(),
                ]
            )
            dataset = torchvision.datasets.MNIST(
                root="data/mnist", train=False, transform=transform, download=True
            )
            # Return first 1/4
            dataset = torch.utils.data.Subset(dataset, range(len(dataset) // 16))

        elif self.hparams.dataset == "cifar10":
            transform = torchvision.transforms.Compose(
                [
                    torchvision.transforms.Resize(self.hparams.sample_size),
                    torchvision.transforms.ToTensor(),
                ]
            )
            dataset = torchvision.datasets.CIFAR10(
                root="data/cifar10", train=False, transform=transform, download=True
            )
            # Return first 1/4
            dataset = torch.utils.data.Subset(dataset, range(len(dataset) // 16))

        elif self.hparams.dataset == "celeba_hq":
            transform = torchvision.transforms.Compose(
                [
                    torchvision.transforms.Resize(self.hparams.sample_size),
                    torchvision.transforms.ToTensor(),
                ]
            )
            dataset = torchvision.datasets.ImageFolder(
                "data/celeba_hq/val", transform=transform
            )
            # Return first 1/4
            dataset = torch.utils.data.Subset(dataset, range(len(dataset) // 4))

        else:
            raise ValueError(f"Unknown dataset: {self.hparams.dataset}")

        return torch.utils.data.DataLoader(
            dataset,
            batch_size=self.hparams.batch_size,
            num_workers=4,
            pin_memory=True,
            persistent_workers=True,
        )



from dataclasses import dataclass

import torch
from torch import nn

__all__ = ["UNetConfig", "UNet"]


@dataclass
class UNetConfig:
    in_channels: int
    out_channels: int
    initial_filters: int = 64


class ConvBlock(nn.Module):
    def __init__(self, in_channels, out_channels):
        super().__init__()
        self.block = nn.Sequential(
            nn.Conv2d(in_channels, out_channels, kernel_size=3, padding=1),
            nn.ReLU(inplace=True),
            nn.Conv2d(out_channels, out_channels, kernel_size=3, padding=1),
            nn.ReLU(inplace=True),
        )

    def forward(self, x):
        return self.block(x)


class UpConvBlock(nn.Module):
    def __init__(self, in_channels, out_channels):
        super().__init__()
        self.upconv = nn.ConvTranspose2d(
            in_channels, out_channels, kernel_size=2, stride=2
        )
        self.conv_block = ConvBlock(in_channels, out_channels)

    def forward(self, x1, x2):
        x1 = self.upconv(x1)
        x = torch.cat([x1, x2], dim=1)
        return self.conv_block(x)


class UNet(nn.Module):
    def __init__(self, config: UNetConfig):
        super().__init__()
        self.config = config

        self.encoder1 = ConvBlock(config.in_channels, config.initial_filters)
        self.encoder2 = ConvBlock(config.initial_filters, config.initial_filters * 2)
        self.encoder3 = ConvBlock(
            config.initial_filters * 2, config.initial_filters * 4
        )
        self.encoder4 = ConvBlock(
            config.initial_filters * 4, config.initial_filters * 8
        )
        self.encoder5 = ConvBlock(
            config.initial_filters * 8, config.initial_filters * 16
        )

        self.pool = nn.MaxPool2d(kernel_size=2, stride=2)

        self.upconv4 = UpConvBlock(
            config.initial_filters * 16, config.initial_filters * 8
        )
        self.upconv3 = UpConvBlock(
            config.initial_filters * 8, config.initial_filters * 4
        )
        self.upconv2 = UpConvBlock(
            config.initial_filters * 4, config.initial_filters * 2
        )
        self.upconv1 = UpConvBlock(config.initial_filters * 2, config.initial_filters)

        self.final_conv = nn.Conv2d(
            config.initial_filters, config.out_channels, kernel_size=1
        )

    def forward(self, x):
        enc1 = self.encoder1(x)
        enc2 = self.encoder2(self.pool(enc1))
        enc3 = self.encoder3(self.pool(enc2))
        enc4 = self.encoder4(self.pool(enc3))
        enc5 = self.encoder5(self.pool(enc4))

        dec4 = self.upconv4(enc5, enc4)
        dec3 = self.upconv3(dec4, enc3)
        dec2 = self.upconv2(dec3, enc2)
        dec1 = self.upconv1(dec2, enc1)

        return self.final_conv(dec1)

    def get_param_count(self):
        return sum(p.numel() for p in self.parameters())


if __name__ == "__main__":
    config = UNetConfig(in_channels=1, out_channels=1, initial_filters=2)

    model = UNet(config)
    x = torch.randn(1, 1, 256, 256)
    preds = model(x)
    print(preds.shape)  # Should be [1, 1, 256, 256]
    print(model.get_param_count())



import torch
from nn_zoo.models.components import (
    ResidualBasicBlock,
    ResidualBottleNeckBlock,
    ResidualStack,
)


def test_residual_basic_block():
    x = torch.randn(1, 3, 32, 32)
    block = ResidualBasicBlock(3, 64, 3)
    assert block(x).shape == (1, 64, 32, 32)


def test_residual_bottle_neck_block():
    x = torch.randn(1, 3, 32, 32)
    block = ResidualBottleNeckBlock(3, 64, 3)
    assert block(x).shape == (1, 64, 32, 32)


def test_residual_stack():
    x = torch.randn(1, 3, 32, 32)
    stack = ResidualStack(3, 3, 64, 3)
    assert stack(x).shape == (1, 64, 32, 32)


if __name__ == "__main__":
    test_residual_basic_block()
    test_residual_bottle_neck_block()
    test_residual_stack()



import torch
from nn_zoo.models import GPTConfig, GPT


def test_gpt():
    config = GPTConfig(
        vocab_size=512,
        block_size=32,
        emb_size=32,
        heads=2,
        num_layers=2,
        attn_dropout=0.1,
        ff_mult=4,
        ff_dropout=0.1,
    )
    model = GPT(config)
    assert model

    x = torch.randint(0, 512, (1, 32))
    y = model(x)
    print(y.shape)
    assert y.shape == (1, 32, 512)


if __name__ == "__main__":
    test_gpt



import torch
from nn_zoo.models import UNet, UNetConfig


def test_unet():
    config = UNetConfig(
        in_channels=3,
        out_channels=3,
        initial_filters=32,
    )
    model = UNet(config)
    assert model

    x = torch.rand(1, 3, 32, 32)
    y = model(x)
    assert y.shape == (1, 3, 32, 32)
